package com.sourcelabs.zipper.zip;

import com.sourcelabs.zipper.InvalidZipFormatException;
import com.sourcelabs.zipper.zip.ZipFile;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.channels.FileChannel;
import java.io.IOException;

/**
 * EndOfCentralDir
 *
 * This class represents the end of the Central Directory for a zip file.
 *
 * UNDONE(willpugh) -- We don't currently deal with multiple disks
 * UNDONE(willpugh) -- Not sure if there is a better way to find the beginning
 * of the EndOfCentralDir record in the case that there is a zip comment
 *
 * @author willpugh  Mar 4, 2006 3:20:58 PM
 */
public class EndOfCentralDir {

    static public final int SIZE_OF_EOCD = 22;

    short       diskNumber;
    short       centralDirDiskNumber;
    short       numEntriesOnThisDisk;
    short       numEntries;
    int         size;
    int         centralDirStart;
    String      zipComment;

    public EndOfCentralDir() {
    }

    public EndOfCentralDir(short diskNumber, short centralDirDiskNumber, short numEntriesOnThisDisk, short numEntries, int size, int centralDirStart, String zipComment) {
        this.diskNumber = diskNumber;
        this.centralDirDiskNumber = centralDirDiskNumber;
        this.numEntriesOnThisDisk = numEntriesOnThisDisk;
        this.numEntries = numEntries;
        this.size = size;
        this.centralDirStart = centralDirStart;
        this.zipComment = zipComment;
    }

    public int load(ByteBuffer buff) throws InvalidZipFormatException {
        return load(buff, 0);
    }


    /**
     * load
     *
     * This function will find the End Of Central Directory from the buffer.  There has
     * to be a better way of doing this, but I don't know what it is.  What I do is:
     *   1)  Count backwards until I find a field that could be a valid zip comment length.
     *   2)  Then verify the signature is correct
     *
     * Since Zip comments are somewhat rare, in the common case I should get this correct the
     * very first time.
     *
     * @param buff This is the Byte Buffer that contains the Central Directory.  If you are chunking
     * your zip file, and aren't loading in the entire file into one buffer, this should buffer should contain
     * the last part of the file.  If this is a multi-disk zip, it should contain the last part of the last file.
     * @param endOffset  This is the offset from the end to start looking.  99.9% of the time, this should be 0.
     * the only time this should have a value is if an earlier call to load returns an incorrect set of information
     * and so you need to re-try.  At which point you should pass back in the value returned.
     * @return the number of bytes from the end of the file that the EOCD ends at.
     * @throws InvalidZipFormatException
     */
    public int load(ByteBuffer buff, int endOffset) throws InvalidZipFormatException {

        int end = buff.limit();

        while (true) {
            short commentLen;

            //First make sure that we get a comment length that makes sense.
            do {
                commentLen = buff.getShort(end-endOffset-2);
            } while ((commentLen != endOffset++) && (endOffset < end));

            if (endOffset == end) {
                throw new InvalidZipFormatException("ERROR:  Could not find the End Of Central Directory. " +
                        " Check to make sure your file is actually a zip file.");
            }

            //Now, check to see if the signature looks right.
            //the beginning is the current position where we found the rigth looking comment length
            //minus the size of the EOCD structure plus 2 (because the current position is not the end of the
            // structure, but instead the beginning of the comment length member)
            int     beginning = end-endOffset-1-SIZE_OF_EOCD+2;
            int     oldPos = buff.position();
            try {
                buff.position(beginning);
                int sig = buff.getInt();
                if (sig == ZipFile.END_OF_CENTRAL_DIRECTORY_SIGNATURE) {
                    diskNumber = buff.getShort();
                    centralDirDiskNumber = buff.getShort();
                    numEntriesOnThisDisk = buff.getShort();
                    numEntries = buff.getShort();
                    size = buff.getInt();
                    centralDirStart = buff.getInt();
                    commentLen = buff.getShort();

                    if (commentLen > 0) {
                        byte[]  commentBytes = new byte[commentLen];
                        buff.get(commentBytes);
                        zipComment = new String(commentBytes);
                    }
                    break;
                }

            } finally {
                buff.position(oldPos);
            }

        }

        return endOffset;
    }

    public int writeToChannel(FileChannel fc) throws IOException {
        //UNDONE(willpugh) -- deal with encodings
        byte[]  rawComment = (zipComment != null ? zipComment.getBytes() : null);
        int     rawCommentSize = (rawComment != null ? rawComment.length : 0);
        int     headerSize = SIZE_OF_EOCD + rawCommentSize;

        ByteBuffer  bb = ByteBuffer.allocate(headerSize);
        bb.order(ByteOrder.LITTLE_ENDIAN);
        bb.putInt(ZipFile.END_OF_CENTRAL_DIRECTORY_SIGNATURE);

        //Disk Number
        bb.putShort(diskNumber);

        //Disk Number that the beginning of the Central Directory is on
        bb.putShort(centralDirDiskNumber);

        //Total number of Central Directory Entries on this disk
        bb.putShort(numEntriesOnThisDisk);

        //Total number of Central Directory Entries
        bb.putShort(numEntries);

        //Size of Central Directory
        bb.putInt(size);

        //Offset of Central Directory
        bb.putInt(centralDirStart);

        bb.putShort((short)rawCommentSize);
        if (rawCommentSize > 0)
            bb.put(rawComment);
        bb.position(0);
        fc.write(bb);
        return headerSize;
    }

}
