package com.sourcelabs.zipper.zip;

import com.sourcelabs.zipper.*;

import java.io.*;
import java.nio.channels.FileChannel;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.*;
import java.util.zip.CRC32;
import java.util.zip.Deflater;

import org.apache.commons.io.IOUtils;

/**
 * Subclass of {@link ZipFile} that allows modification of the zip.
 *
 * The idea here is that you build up your list of changes calling some variation of AddEntry, ModifyEntry,
 * AddOrModifyEntry or RemoveEntry and then call ApplyChanges to have them
 * to have the changes commited to the zip file.
 *
 * For convienence, all the APIs that allow you to add or modify an entry allow four different
 * forms for supplying the new content:
 *
 * <ul>
 *   <li>{@link java.io.File} - At insertion time, the contents of this file will be put into the zip entry</li>
 *   <li>{@link java.io.InputStream} - The content will be pulled from this stream to put in the zip entry</li>
 *   <li>byte[] - The content in this byte array will be put in the zip stream</li>
 *   <li>{@link ZipCallback} - For supplying content via a callback, you need not call AddEntry, etc. with a
 * second parameter, however, you <em>MUST</em> call {@link #ApplyChanges(ZipCallback)} to apply the changes.
 * For each file, your provided callback will be called, and it will be up to you to determine what content be
 * put in the entry.</li>
 * </ul>
 *
 * This implementation currently needs JDK 1.4 or above, because it uses NIO + Direct Buffers.
 *
 * @author willpugh  Apr 6, 2006 11:03:38 PM
 */
public class ModifiableZipFile extends ZipFile {

    public static final Object DELETE_TOKEN = new Object();
    public static final Object CALLBACK_TOKEN = new Object();

    int         numDeleted = 0;
    int         numChanged = 0;
    LinkedList  addedEntries = new LinkedList();
    HashMap     changedEntries = new HashMap();


    /**
     * Class constructor specifying the name of the zipfile.
     *
     * @param zipName The name of the zipfile to create/modify
     * @throws IOException
     */
    public ModifiableZipFile(String zipName) throws IOException {
        super(zipName);
    }

    /**
     * Class contructor specifying the file to open or create.
     *
     * @param zipFile The file to open or create
     */
    public ModifiableZipFile(File zipFile) {
        super(zipFile);
    }

    protected String getOpenMode() {
        return "rw";
    }


    /**
     * Adds an entry to be added when ApplyChanges is called.
     *
     * @param entry The entry to add
     * @param fileContents The file to use for the contents of the new entry
     * @throws EntryAlreadyExistsException if the entry already exists, this runtime exception will be thrown.
     */
    public void AddEntry(ZipEntry entry, File fileContents) {
        _addEntry(entry, fileContents, true);
    }

    /**
     * Adds an entry to be added when ApplyChanges is called.
     *
     * @param entry The entry to add
     * @param streamContent This inputstream will be used for the contents of the new entry
     * @throws EntryAlreadyExistsException if the entry already exists, this runtime exception will be thrown.
     */
    public void AddEntry(ZipEntry entry, InputStream streamContent) {
        _addEntry(entry, streamContent, true);
    }

    /**
     * Adds an entry to be added when ApplyChanges is called.
     *
     * @param entry The entry to add
     * @param byteContent this byte array will e used for the contents of the new entry
     * @throws EntryAlreadyExistsException if the entry already exists, this runtime exception will be thrown.
     */
    public void AddEntry(ZipEntry entry, byte[] byteContent) {
        _addEntry(entry, byteContent, true);
    }

    /**
     * Adds an entry to be added when ApplyChanges is called.  Since, no content
     * is provided immediately, {@link #ApplyChanges(ZipCallback)} must be called with a
     * callback.  The callback will be responsible for setting the content appropriately.
     *
     * @param entry The entry to add
     * @throws EntryAlreadyExistsException if the entry already exists, this runtime exception will be thrown.
     */
    public void AddEntry(ZipEntry entry) {
        _addEntry(entry, CALLBACK_TOKEN, true);
    }

    /**
     * Does the actual work for adding entries to the modification list.  This is
     * used by both AddEntry and ModifyEntry with the only difference being what is sent in
     * for <code>isAdd</code>
     *
     * @param entry new/modified entry
     * @param o content or <code>CALLBACK_TOKEN</code>
     * @param isAdd whether this is an Add or Modification operation
     * @throws EntryAlreadyExistsException this is an Add operation and the entry already exists
     * @throws EntryDoesNotExistException this is a modification operation and the entry does not exist
     */
    protected void _addEntry(ZipEntry entry, Object o, boolean isAdd) {

        try {
            if (!index.isLoaded()) {
                Init();
                index.load(bb);
            }
            boolean entryExists = index.containsName(entry.getName()) ||
                    changedEntries.containsKey(entry.getName());
            if (entryExists && isAdd) {
                throw new EntryAlreadyExistsException("Cannot add zip entry, " + entry.getName() + " already exists.");
            } else if (!entryExists && !isAdd) {
                throw new EntryDoesNotExistException("Cannot modify zip entry, " + entry.getName() + " since it does not exists.");
            }
            changedEntries.put(entry.getName(), new EntryHolder(entry, o));
            if (!isAdd)
                numChanged++;

        } catch (InvalidZipFormatException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }

    }


    /**
     * Modifies an entry in the zip file.  The contents provided by this call will
     * completely replace the current contents of the existing zip entry.
     *
     * @param entry the entry to replace
     * @param fileContents the file to get the new contents from
     * @throws EntryDoesNotExistException entry does not exist to modify.  This is a runtime exception
     */
    public void ModifyEntry(ZipEntry entry, File fileContents) {
        _addEntry(entry, fileContents, false);
    }

    /**
     * Modifies an entry in the zip file.  The contents provided by this call will
     * completely replace the current contents of the existing zip entry.
     *
     * @param entry the entry to replace
     * @param streamContent the {@link java.io.InputStream} to get the new contents from
     * @throws EntryDoesNotExistException entry does not exist to modify.  This is a runtime exception
     */
    public void ModifyEntry(ZipEntry entry, InputStream streamContent) {
        _addEntry(entry, streamContent, false);
    }

    /**
     * Modifies an entry in the zip file.  The contents provided by this call will
     * completely replace the current contents of the existing zip entry.
     *
     * @param entry the entry to replace
     * @param byteContent the byte array to get the new contents from
     * @throws EntryDoesNotExistException entry does not exist to modify.  This is a runtime exception
     */
    public void ModifyEntry(ZipEntry entry, byte[] byteContent) {
        _addEntry(entry, byteContent, false);
    }

    /**
     * Modifies an entry in the zip file.  The contents provided by this call will
     * completely replace the current contents of the existing zip entry.  Since, there
     * is no content provided with this call, {@link #ApplyChanges(ZipCallback)} must be
     * called with a callback that will provide the appropriate content.
     *
     * @param entry the entry to replace
     * @throws EntryDoesNotExistException entry does not exist to modify.  This is a runtime exception
     */
    public void ModifyEntry(ZipEntry entry) {
       _addEntry(entry, CALLBACK_TOKEN, false);
    }

    /**
     * Adds or Modifies an entry in the zip file.  The contents provided by this call will
     * completely replace the current contents of the existing zip entry.
     *
     * @param entry the entry to replace
     * @param fileContents the file to get the new contents from
     */
    public void AddOrModifyEntry(ZipEntry entry, File fileContents) {
        changedEntries.put(entry.getName(), new EntryHolder(entry, fileContents));
    }

    /**
     * Adds or modifies an entry in the zip file.  The contents provided by this call will
     * completely replace the current contents of the existing zip entry.
     *
     * @param entry the entry to replace
     * @param streamContent the {@link java.io.InputStream} to get the new contents from
     */
    public void AddOrModifyEntry(ZipEntry entry, InputStream streamContent) {
        changedEntries.put(entry.getName(), new EntryHolder(entry, streamContent));
    }

    /**
     * Adds or modifies an entry in the zip file.  The contents provided by this call will
     * completely replace the current contents of the existing zip entry.
     *
     * @param entry the entry to replace
     * @param byteContent the byte array to get the new contents from
     * @throws EntryDoesNotExistException entry does not exist to modify.  This is a runtime exception
     */
    public void AddOrModifyEntry(ZipEntry entry, byte[] byteContent) {
        changedEntries.put(entry.getName(), new EntryHolder(entry, byteContent));
    }

    /**
     * Adds or modifies an entry in the zip file.  The contents provided by this call will
     * completely replace the current contents of the existing zip entry.  Since, there
     * is no content provided with this call, {@link #ApplyChanges(ZipCallback)} must be
     * called with a callback that will provide the appropriate content.
     *
     * @param entry the entry to replace
     */
    public void AddOrModifyEntry(ZipEntry entry) {
       changedEntries.put(entry.getName(), new EntryHolder(entry, CALLBACK_TOKEN));
    }

    /**
     * Removes the entry with the supplied name from the zip file.  The entry will
     * not actually be removed until ApplyChanges is called.
     *
     * @param name the name of the entry to be removed
     */
    public void RemoveEntry(String name) {
        changedEntries.put(name, DELETE_TOKEN);
        numDeleted++;
    }

    /**
     * applies the changes recorded by calls to AddEntry, ModifyEntry, AddOrModifyEntry or DeleteEntry.
     *
     * This will run modify the bytes in the existing zipfile to apply those changes.  Until this point,
     * none of the changes called via the Add, Modify, Delete methods will be noticed by calles of the
     * zip api, and the changes will not be written to disk.
     *
     * @return size in bytes of the changed zip file
     * @throws IOException
     * @throws InvalidZipFormatException
     */
    public int ApplyChanges() throws IOException, InvalidZipFormatException {
       return ApplyChanges(null);
    }

    /**
     * applies the changes recorded by calls to AddEntry, ModifyEntry, AddOrModifyEntry or DeleteEntry.
     *
     * This will run modify the bytes in the existing zipfile to apply those changes.  Until this point,
     * none of the changes called via the Add, Modify, Delete methods will be noticed by calles of the
     * zip api, and the changes will not be written to disk.
     *
     * @param callback callback that occur on files being written to disk.  This will only call back for
     * new or modified files.
     * @return size in bytes of the changed zip file
     * @throws IOException
     * @throws InvalidZipFormatException
     */
    public int ApplyChanges(ZipCallback callback) throws IOException, InvalidZipFormatException {

        //UNDONE(willpugh) -- Currently, we don't do any file locks for this.  Maybe we should?
        //The problem is not sure whether it will stat getting annoying about leaving locks around, etc.
        //Maybe locks should be optional?

        int         bytesWritten = 0;
        ArrayList   entriesToKeep = new ArrayList(index.getCount() - numChanged);
        bytesWritten += WriteExistingEntries(fc, entriesToKeep);
        fc.position(bytesWritten);


        //
        //Run through new and modified entries
        ZipIndexBase    newIndex = new ZipIndexBase();
        ArrayList       newEntries = new ArrayList(changedEntries.size() - numDeleted);

        newIndex.init(changedEntries.size() - numDeleted);
        bytesWritten += WriteModifiedEntries(fc, newIndex, newEntries, callback);


        //Add Central Directory
        fc.position(bytesWritten);
        bytesWritten += WriteCentralDirectories(fc, new ZipIndexBase[] {index, newIndex}, new ArrayList[] {entriesToKeep, newEntries});
        assert(bytesWritten == fc.position());

        //Clear modified files list
        changedEntries.clear();
        numDeleted = 0;
        numChanged = 0;

        fc.truncate(bytesWritten);
        fc.force(true);

        //Now, we need to remap the memory mapped file and go
        bb = mapFile(bytesWritten);
        return bytesWritten;
    }

    /**
     * Writes the entries that already exist in the zip file, and are not being modified or deleted.
     * We do this by finding the blocks of entries that are unchanged, and slide those up to eliminat
     * the entries that we are either deleting or modifying.  Then we change the indexs in our ZipIndex
     * so we know their new offsets.
     *
     * @param fc the file channel everything is written to
     * @param entriesToKeep an arraylist that will be filled in with the entries that are keep in this block.
     * @return the number of bytes for the section of the zip that contains the unchanged entries.
     * @throws IOException
     */
    protected int WriteExistingEntries(FileChannel fc, ArrayList entriesToKeep) throws IOException {
        if (!index.isLoaded()) {
            index.load(bb);
        }

        int eliminationStart = -1;
        int keepersStart = -1;
        int offset = 0;
        int bytesWritten = 0;


        //First run through zip file, removing entries that are changing
        //We will compact the zip as we go, but will need to keep track of
        //how big the central directory is after removing all the entries
        //we need to remove
        for (int i=0; i<index.getCount(); ) {

            //This begins a block of entries we want to get rid of.  Run to the end,
            //and mark the beginning of the block to remove
            if (changedEntries.containsKey(index.getName(i))) {
                //Set the beginning of the block to eliminate
                eliminationStart = index.getOffset(i) - offset;

                //Look for the end of the block
                do {
                    offset += index.getFileEntrySize(i);
                    index.setOffset(i, -1);
                    i++;
                } while ((i < index.getCount()) && (changedEntries.containsKey(index.getName(i))));

            } else {
                //This marks a block of entries we want to keep.  Run to the end
                //and then move the block the appropriate distance.
                int sizeOfBlock = 0;
                keepersStart = index.getOffset(i);

                //Run through all the blocks we want to keep
                do {
                    entriesToKeep.add(index.getEntry(i, bb));
                    sizeOfBlock += index.getFileEntrySize(i);
                    index.setOffset(i, index.getOffset(i) - offset);
                    i++;
                } while ((i < index.getCount()) && (!changedEntries.containsKey(index.getName(i))));

                if (offset != 0) {
                    //Do the shift

                    //UNDONE(willpugh) -- Is there a better way to do this without an intermediate copy?
                    //  I ran into problems trying TransferFrom, but I'll bet I can come up with a better
                    // way.  Is there something in Commons IO?
                    ByteBuffer  bbTemp = ByteBuffer.allocate(sizeOfBlock);
                    bbTemp.order(ByteOrder.LITTLE_ENDIAN);
                    fc.read(bbTemp, keepersStart);
                    bbTemp.position(0);
                    bytesWritten += fc.write(bbTemp, eliminationStart);

                    //bytesWritten += (int)fc.position(eliminationStart).transferFrom(fc, keepersStart, sizeOfBlock);
                } else {
                    bytesWritten += sizeOfBlock;
                }
            }
        }
        return bytesWritten;
    }

    /**
     * Writes the entries that are either new or modified.  For these entries we need to go through
     * the whole shabang of compressing, getting the CRCs, doing the callbacks, etc.
     *
     * @param fc the file channel to write to
     * @param index the new index we are creating with that will contain all the new and modified entries
     * @param entries an ArrayList of the new/modified entries
     * @param callback the callback to get content changes
     * @return the size of this section of the zip file
     * @throws IOException
     */
    protected int WriteModifiedEntries(FileChannel fc, ZipIndexBase index, ArrayList entries, ZipCallback callback) throws IOException {
        int             bytesWritten = 0;
        int             nextPosition = (int) fc.position();

        int i =0;
        Iterator  changedIter = changedEntries.entrySet().iterator();
        while (changedIter.hasNext()) {

            Map.Entry   currEntry = (Map.Entry) changedIter.next();
            if (currEntry.getValue() == DELETE_TOKEN)
                continue;

            EntryHolder     entryHolder = (EntryHolder) currEntry.getValue();
            ZipEntry currZipEntry = entryHolder.entry;
            Object value = entryHolder.value;

            //Set the position to the beginning of the new block
            fc.position(nextPosition);
            int headerSize = currZipEntry.writeToChannel(fc);

            int             retVal = ZipCallback.PROCESS_ENTRY;
            InputStream     is = null;

            //UNDONE(willpugh) -- look at header and use the compression type that is set there.
            ZipFileStream   zfs = new ZipFileStream(fc);

            if (value instanceof File) {
                is = new FileInputStream((File) value);
            } else if (value instanceof byte[]) {
                is = new ByteArrayInputStream((byte[])value);
            } else if (value instanceof InputStream) {
                is = (InputStream) value;
            }

            if (callback != null) {
                retVal = callback.addEntry(currZipEntry, is, zfs);
            } else if (is == null) {
                assert(value == CALLBACK_TOKEN);
                throw new ZipException("Cannot add entry without a body, and then not provide a callback to ApplyChanges.");
            }

            if (retVal == ZipCallback.REMOVE_ENTRY) {

                //Invalidate section of i and
                //Keep beginning of Block the same, so this gets written over
                index.setOffset(i++, -1);
                continue;
            } else if (retVal == ZipCallback.PROCESS_ENTRY) {
                //Do the copy ourselves
                IOUtils.copy(is, zfs);
            }

            zfs.close();

            bytesWritten += (headerSize + zfs.bytesWritten);

            entries.add(currZipEntry);
            fc.position(nextPosition);
            currZipEntry.fixupFileParts(fc, (int) zfs.crc.getValue(),
                                        zfs.bytesWritten, zfs.uncompressdedBytesWritten);
            index.setName(i, currZipEntry.getName());
            index.setCompressedSize(i, currZipEntry.getCompressedSize());
            index.setUncompressedSizes(i, (int) currZipEntry.getUncompressedSize());
            index.setOffset(i, nextPosition);
            i++;

            nextPosition += headerSize + zfs.bytesWritten;
        }
        return bytesWritten;
    }

    /**
     * Takes the lists of entries and the list of indices and writes them out as the Central Directory
     * at the end of the file.
     *
     * @param fc FileChannel to write to
     * @param indexes an array of the idexes that contain offset informtion
     * @param entries lists of the entries that contain name, compression type, etc. information
     * @return the size of the Central Directory in bytes
     * @throws IOException
     */
    protected int WriteCentralDirectories(FileChannel fc, ZipIndexBase[] indexes, ArrayList[] entries) throws IOException {
        int     cdOffset = (int) fc.position();
        int     numEntries = 0;
        int     sizeOfDirectory = 0;

        for (int i=0; i<indexes.length; i++) {

            int             entriesInIndex = 0;
            ZipIndexBase    currIndex = indexes[i];
            ArrayList       currEntryList = entries[i];

            for (int j=0; j < currIndex.getCount(); j++) {
                if (currIndex.getOffset(j) == -1)
                    continue;

                sizeOfDirectory += currIndex.writeToChannel(j, fc, (ZipEntry) currEntryList.get(entriesInIndex));
                entriesInIndex++;
            }
            numEntries += entriesInIndex;
        }

        //Merge the Central Directories
        index = MergeCentralDirectories(indexes, numEntries, cdOffset);

        //Write the End Of Central Directory
        EndOfCentralDir eocd = new EndOfCentralDir((short)0, (short)0, (short)numEntries, (short)numEntries, sizeOfDirectory, cdOffset, zipComment);

        int bytesWritten = sizeOfDirectory;
        bytesWritten += eocd.writeToChannel(fc);
        return bytesWritten;
    }

    /**
     * Merges the different indices to create a single ZipINdex that this ZipFile can use for additional
     * usage.  THis way we don't need to reload the file or anything crazy like that.
     *
     * @param indexes the ZipIndexes that contain all the offset information for the entries
     * @param numEntries the number of the entries in the Central Directory
     * @param cdOffset the offset in the file to the beginning of the Central Directory
     * @return The final ZipIndex this ZipFile can use
     */
    protected ZipIndex MergeCentralDirectories(ZipIndexBase[] indexes, int numEntries, int cdOffset) {

        ZipIndex    zipIndex = new ZipIndex();
        zipIndex.init(numEntries);
        zipIndex.setCentralDirStart(cdOffset);

        int currEntries = 0;
        for (int i=0; i<indexes.length; i++) {
            ZipIndexBase    currIndex = indexes[i];
            for (int j=0; j<currIndex.getCount(); j++) {

                if (currIndex.getOffset(j) == -1)
                    continue;

                zipIndex.setName(currEntries, currIndex.getName(j));
                zipIndex.setOffset(currEntries, currIndex.getOffset(j));
                zipIndex.setCompressedSize(currEntries, currIndex.getCompressedSize(j));
                zipIndex.setUncompressedSizes(currEntries, currIndex.getUncompressedSizes(j));

                String comment = currIndex.getComment(j);
                if (comment != null && comment.length() > 0) {
                    zipIndex.addComment(currEntries, comment);
                }
                currEntries++;
            }
        }
        assert(currEntries == numEntries);
        zipIndex.setLoaded(true);
        return zipIndex;
    }


    public static final int DEFAULT_BUFF_SIZE = 4096;

    /**
     * A stream that will compress input data, as well as compute the CRC32 for it.
     */
    static public class ZipFileStream extends OutputStream {
        FileChannel     fc;
        CRC32           crc = new CRC32();
        Deflater        deflater;

        //Uncompressed Data Management
        int             bytesToWrite = 0;
        int             uncompressdedBytesWritten = 0;
        BuffHolder      currBuff = null;

        //Compressed Data Management
        int             buffLen;
        int             bytesWritten;
        byte            outBuff[];
        //ByteBuffer      byteBuff;

        public ZipFileStream(FileChannel fc) {
            this(fc, Deflater.DEFAULT_COMPRESSION);
        }

        public ZipFileStream(FileChannel fc, int deflaterLevel) {
            this(fc, deflaterLevel, DEFAULT_BUFF_SIZE);
        }

        public ZipFileStream(FileChannel fc, int deflaterLevel, int buffLen) {
            this.fc = fc;
            this.buffLen = buffLen;
            this.outBuff = new byte[buffLen];
            deflater = new Deflater(deflaterLevel, true);

        }

        public void write(int i) throws IOException {
            if (fc==null) {
                throw new IOException("Stream is closed");
            }

            if (currBuff == null) {
                currBuff = new BuffHolder(new byte[buffLen]);
            }
            currBuff.write(i);
            if (++bytesToWrite >= buffLen) {
                writeBuff(currBuff);
            }

        }

        public void write(byte[] bytes) throws IOException {
            if (currBuff != null)
                writeBuff(currBuff);

            writeBuff(new BuffHolder(bytes));
        }

        public void write(byte[] bytes, int i, int i1) throws IOException {
            if (currBuff != null)
                writeBuff(currBuff);

            writeBuff(new BuffHolder(bytes, i, i1));
        }

        public void flush() throws IOException {
            if (currBuff != null)
                writeBuff(currBuff);
        }

        public void close() throws IOException {
            flush();
            deflater.finish();
            while (!deflater.finished()) {
                int bytesDeflated = deflater.deflate(outBuff);
                if (bytesDeflated > 0) {
                    ByteBuffer byteBuff = ByteBuffer.wrap(outBuff, 0, bytesDeflated);
                    byteBuff.order(ByteOrder.LITTLE_ENDIAN);

                    fc.write(byteBuff);
                    bytesWritten += bytesDeflated;
                } else {
                    break;
                }
            }
            bytesToWrite = -1;
            //crc.reset();
            
            //Don't close the FileChannel, other will keep using it.
            fc = null;
        }


        protected void writeBuff(BuffHolder buff) throws IOException {
            if (fc==null) {
                throw new IOException("Stream is closed");
            }

            if (buff.num > 0) {
                uncompressdedBytesWritten += buff.num;
                deflater.setInput(buff.bytes, buff.startIndex, buff.num);
                crc.update(buff.bytes, buff.startIndex, buff.num);
                drainDeflater();
            }
        }

        protected void drainDeflater() throws IOException {

            while (!deflater.needsInput()) {
                int bytesDeflated = deflater.deflate(outBuff);
                if (bytesDeflated > 0) {
                    ByteBuffer byteBuff = ByteBuffer.wrap(outBuff, 0, bytesDeflated);
                    byteBuff.order(ByteOrder.LITTLE_ENDIAN);

                    fc.write(byteBuff);
                    bytesWritten += bytesDeflated;
                } else {
                    break;
                }
            }

        }

    }

    /**
     * A dorky helper class that keeps track of what bytes in a given
     * byte array are the ones we want to write.
     */
    public static final class BuffHolder {
        public byte[]   bytes;
        public int      startIndex;
        public int      num;

        public BuffHolder(byte[] bytes) {
            this(bytes, 0, 0);
        }

        public BuffHolder(byte[] bytes, int startOffset) {
            this(bytes, startOffset, 0);
        }

        public BuffHolder(byte[] bytes, int startOffset, int len) {
            this.bytes = bytes;
            this.startIndex = startOffset;
            this.num = len;
        }

        public boolean isFull() {
            return bytes.length <= (startIndex + num);
        }


        public void write(int i) {
            bytes[startIndex + num++] = (byte)i;
        }
    }

    /**
     * A dorky helper class to match ZipEntries to the content values that will
     * be used to add or modify them.
     */
    public static class EntryHolder {
        ZipEntry    entry;
        Object      value;

        public EntryHolder(ZipEntry entry, Object value) {
            this.entry = entry;
            this.value = value;
        }
    }

}
