package com.sourcelabs.zipper.zip;

import com.sourcelabs.zipper.SignatureMismatchException;
import com.sourcelabs.zipper.zip.ZipFile;
import com.sourcelabs.zipper.ZipBodyInputStream;
import com.sourcelabs.zipper.PayloadInputStream;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.channels.FileChannel;
import java.io.InputStream;
import java.io.IOException;
import java.util.GregorianCalendar;
import java.util.Calendar;

/**
 * Implements all the guts of loading and saving a ZipEntry to the zipfile.
 *
 * This includes all the nasty bits like converting Java Dates/Times to the Dates/Times
 * used in zips.
 *
 * @author willpugh  Mar 3, 2006 11:21:57 PM
 */
public class ZipEntryImpl {

    public static final int HEADER_SIZE = 30;
    private static final int LENGTH_AT_END = 0x00000004;

    private static final short COMPRESSION_STORED = 0;
    private static final short COMPRESSION_DEFLATED = 8;
    private static final short PKVERSION_2_0 = 20;



    private short               version = PKVERSION_2_0;
    private short               flags;
    private short               compression = COMPRESSION_DEFLATED;
    int                 rawtime;
    private int                 crc_32;
    private int                 compressedSize;
    private int                 uncompressedSize;
    private int                 dataPos;

    private String              name;
    private byte[]              extra;

    public ZipEntryImpl() {
    }

    public ZipEntryImpl(String name) {
        this.name = name;
    }

    public ZipEntryImpl(ZipEntryImpl h) {
        this.version = h.version;
        this.flags = h.flags;
        this.compression = h.compression;
        this.setRawtime(h.getRawtime());
        this.crc_32 = h.crc_32;
        this.compressedSize = h.compressedSize;
        this.uncompressedSize = h.uncompressedSize;
        this.dataPos = h.dataPos;
        this.name = h.name;
        this.extra = h.extra;
    }

    /**
     * Loads an entry from a ByteBuffer containing that part of the
     * ZipFile.  The ByteBuffer must have it's position set to the beginning of the
     * entry to load.
     * @param bb
     */
    public void loadFromBuffer(ByteBuffer bb) {
        version = bb.getShort();
        flags = bb.getShort();
        compression = bb.getShort();

        rawtime = bb.getInt();
        crc_32 = bb.getInt();
        compressedSize = bb.getInt();
        uncompressedSize = bb.getInt();

        short   fileNamelen = bb.getShort();
        short   extraLen = bb.getShort();

        byte[]  rawName = new byte[fileNamelen];
        byte[]  extra = new byte[extraLen];
        bb.get(rawName);
        bb.get(extra);
        dataPos = bb.position();

        name = new String(rawName);
        this.extra = extra;
    }

    /**
     * Writes this entrie's local header information to the file channel.  This will
     * not write any of the contents.
     * @param fc the filechalle to write to.  The position must be set to the place the
     * entry should be written to.
     * @return the size in bytes of the entry header
     * @throws IOException
     */
    public int writeToChannel(FileChannel fc) throws IOException {

        //UNDONE(willpugh) -- Deal with encodings.  Default should be UTF-8, but folks
        //shoudl be able to set this.
        byte[]  rawName = name.getBytes();
        int     extraSize = (extra != null ? extra.length : 0);
        int     headerSize = HEADER_SIZE + rawName.length + extraSize;

        ByteBuffer  buf = ByteBuffer.allocate(headerSize);
        buf.order(ByteOrder.LITTLE_ENDIAN);
        buf.putInt(ZipFile.FILE_HEADER_SIGNATURE);
        buf.putShort(version);
        buf.putShort(flags);
        buf.putShort(compression);
        buf.putInt(getRawtime());
        buf.putInt(crc_32);
        buf.putInt(compressedSize);
        buf.putInt(uncompressedSize);

        buf.putShort((short) rawName.length);
        buf.putShort((short) extraSize);
        buf.put(rawName);
        if (extra != null)
            buf.put(extra);

        buf.position(0);
        fc.write(buf);
        return headerSize;
    }

    /**
     * Fixes up the crc32, the compressed size and the uncompressed size in a zip entrie's local header.
     * Often, we don't actually know this stuff until after we've written the file, so this method is how
     * we go back and fix up what we wrote.
     *
     * @param fc {@link FileChannel} to write to.
     * @param crc_32
     * @param compressedSize
     * @param uncompressedSize
     * @throws IOException
     */
    public void fixupFileParts(FileChannel fc, int crc_32, int compressedSize, int uncompressedSize) throws IOException {

        ByteBuffer  bb = ByteBuffer.allocate(12);
        bb.order(ByteOrder.LITTLE_ENDIAN);
        bb.putInt(crc_32);
        bb.putInt(compressedSize);
        bb.putInt(uncompressedSize);
        fc.position(fc.position() + 14);

        bb.position(0);
        fc.write(bb);

        this.crc_32 = crc_32;
        this.compressedSize = compressedSize;
        this.uncompressedSize = uncompressedSize;
        return;
    }

    /**
     * Loads the entry from teh local header information.  This is the
     * same as {@link #loadFromBuffer(java.nio.ByteBuffer)} but it verifies the
     * first four bytes match the Local File Header Signature.
     * @param bb {#link ByteBuffer} to load from.  The position should be set to the
     * beginning of the local header.
     * @throws SignatureMismatchException
     */
    public void load(ByteBuffer bb) throws SignatureMismatchException {
        int signature = bb.getInt();
        if (signature != ZipFile.FILE_HEADER_SIGNATURE) {
            throw new SignatureMismatchException();
        }
        loadFromBuffer(bb);
    }

    static public final int     YEAR_MASK =     0xFE000000;
    static public final int     MONTH_MASK =    0x01E00000;
    static public final int     DAY_MASK =      0x001F0000;
    static public final int     HOUR_MASK =     0x0000F800;
    static public final int     MINUTE_MASK =   0x000007E0;
    static public final int     SECOND_MASK =   0x0000001F;


    static protected GregorianCalendar convertToJavaTime(int time) {
        int     year = ((time & YEAR_MASK) >>> 25) + 1980;
        int     month = ((time & MONTH_MASK) >>> 21) - 1;
        int     date = (time & DAY_MASK) >>> 16;
        int     hour = (time & HOUR_MASK) >>> 11;
        int     minute = (time & MINUTE_MASK) >>> 5;
        int     second = ((time & SECOND_MASK) << 1);

        return new GregorianCalendar(year, month, date, hour, minute, second);
    }

    public static final int DOS_JAN1_1980 = (int)0x210000;
    static protected int convertToRawTime(long javaTime) {
        GregorianCalendar   cal = new GregorianCalendar();
        cal.setTimeInMillis(javaTime);

        int year = cal.get(Calendar.YEAR) - 1980;
        if (year < 0)
            return DOS_JAN1_1980;

        int     curr = (year << 25);
        curr |= ((cal.get(Calendar.MONTH) + 1) << 21);
        curr |= (cal.get(Calendar.DAY_OF_MONTH) << 16);
        curr |= (cal.get(Calendar.HOUR_OF_DAY) << 11);
        curr |= (cal.get(Calendar.MINUTE) << 5);
        curr |= (cal.get(Calendar.SECOND) >>> 1);
        return curr;
    }


    byte[]  getBody(ByteBuffer buf) {
        byte[]  body = new byte[compressedSize];

        int oldPos = buf.position();
        try {
            buf.get(body);
        } finally {
            buf.position(oldPos);
        }
        return body;
    }

    public InputStream  getBodyStream(ByteBuffer buf) {

        InputStream     retVal = null;
        switch (compression) {
            case    COMPRESSION_STORED:
                retVal = new PayloadInputStream(dataPos, compressedSize, buf);
                break;

            case    COMPRESSION_DEFLATED:
                retVal = new ZipBodyInputStream(dataPos, compressedSize, buf);
                break;
        }

        return retVal; 
    }



    public void load(ByteBuffer bb, int offset, int uncompressLen, int compressedLen) throws SignatureMismatchException {
        int oldpos = bb.position();
        bb.position(offset);
        try {
            load(bb);
            this.compressedSize = compressedLen;
            this.uncompressedSize = uncompressLen;
        } finally {
            bb.position(oldpos);
        }
    }

    public void setTime(long l) {
        setRawtime(convertToRawTime(l));
    }

    public long getTime() {
        return convertToJavaTime(getRawtime()).getTimeInMillis();
    }

    public Calendar getTimeAsCalendar() {
        return convertToJavaTime(getRawtime());
    }

    public boolean isLengthAtEnd() {
        return (flags & LENGTH_AT_END) == LENGTH_AT_END;
    }

    public static int getLengthAtEnd() {
        return LENGTH_AT_END;
    }

    public static short getCompressionStored() {
        return COMPRESSION_STORED;
    }

    public static short getCompressionDeflated() {
        return COMPRESSION_DEFLATED;
    }

    public short getVersion() {
        return version;
    }

    public void setVersion(short version) {
        this.version = version;
    }

    public short getFlags() {
        return flags;
    }

    public void setFlags(short flags) {
        this.flags = flags;
    }

    public short getCompression() {
        return compression;
    }

    public void setCompression(short compression) {
        this.compression = compression;
    }

    public int getCrc_32() {
        return crc_32;
    }

    public void setCrc_32(int crc_32) {
        this.crc_32 = crc_32;
    }

    public int getCompressedSize() {
        return compressedSize;
    }

    public void setCompressedSize(long compressedSize) {
        this.compressedSize = (int) compressedSize;
    }

    public long getUncompressedSize() {
        return uncompressedSize;
    }

    public void setUncompressedSize(int uncompressedSize) {
        this.uncompressedSize = uncompressedSize;
    }

    public int getDataPos() {
        return dataPos;
    }

    public void setDataPos(int dataPos) {
        this.dataPos = dataPos;
    }

    public String getName() {
        return name;
    }

    public void setName(String name) {
        this.name = name;
    }

    public byte[] getExtra() {
        return extra;
    }

    public void setExtra(byte[] extra) {
        this.extra = extra;
    }

    public boolean isDirectory() {
        return name.endsWith("/");
    }

    public int getRawtime() {
        if(rawtime==0) {
            setTime(System.currentTimeMillis());
        }
        return rawtime;
    }

    public void setRawtime(int rawtime) {
        this.rawtime = rawtime;
    }
}
