package com.sourcelabs.zipper.zip;

import com.sourcelabs.zipper.InvalidZipFormatException;
import com.sourcelabs.zipper.ZipVisitor;
import com.sourcelabs.zipper.EntryIterator;
import com.sourcelabs.zipper.IndexIterator;

import java.io.*;
import java.nio.channels.FileChannel;
import java.nio.MappedByteBuffer;
import java.nio.ByteOrder;
import java.util.Iterator;
import java.util.Enumeration;

import org.apache.commons.io.IOUtils;
import org.apache.commons.io.FileUtils;
import org.apache.commons.io.filefilter.IOFileFilter;
import org.apache.commons.io.filefilter.TrueFileFilter;
import org.apache.commons.io.filefilter.FalseFileFilter;

/**
 * Implements a ZipFile.  This is very similar to {@link java.util.zip.ZipFile} except that it
 * has some additional interation routines, and it uses internal structures that we need to be able
 * to implement {@link ModifiableZipFile}.
 *
 * {@link #visitFile(com.sourcelabs.zipper.ZipVisitor)} allows you to interate over the file in a called back manner.
 *
 * In general, this class is not a terribly difficult one to implement.  A couple of the Caveats here:
 *   <ul>
 *     <li>{@link ZipIndex} is the class that reads/writes the Central Directory of the Zip and is used for
 * implementing much of the functionality needed for finding an entry, or figuring out how many are in the zipfile.</li>
 *     <li>{@link ZipEntry} is the class that reads/writes a single entry.  Often, the offset to the entry will be
 * read from a ZipIndex.</li>
 *     <li>Iteration.  We do a sorta dorky thing with zip iteration.  If you try to iterate before doing anything with
 * the zipfile, we will iterate without ever loading the Central Directory.  Need to do perf analysis to see if this
 * is terribly necassary, but it is the reason for there being two different types of iterators :
 * {@link EntryIterator} and {@link IndexIterator}</li>
 *     <li>Currently, JDK 1.4 is required because we use NIO and DirectBuffers.</li> 
 *   </ul>
 *
 * @author willpugh  Mar 3, 2006 11:19:25 PM
 */
public class ZipFile {
    File                        thisFile;
    RandomAccessFile            raf;
    FileChannel                 fc;
    MappedByteBuffer            bb;
    ZipIndex                    index = new ZipIndex();
    String                      zipComment;

    /**
     * constructor for ZipFile that takes the name of the file to open.
     * @param fileName teh name of the file to open
     * @throws IOException
     */
    public ZipFile(String fileName) throws IOException {
        thisFile = new File(fileName);
        
    }

    /**
     * contructor for ZipFile that takes the file to open.
     * @param file the file to open as a zip
     */
    public ZipFile(File file) {
        thisFile = file;

    }

    public static final int FILE_HEADER_SIGNATURE = 0x04034b50;
    public static final int EXTRA_DATA_SIGNATURE = 0x08064b50;
    public static final int CENTRAL_DIRECTORY_SIGNATURE = 0x02014b50;
    public static final int DIGITAL_SIGNATURE_SIGNATURE = 0x05054b50;
    public static final int END_OF_CENTRAL_DIRECTORY_64_SIGNATURE = 0x06064b50;
    public static final int END_OF_CENTRAL_DIRECTORY_LOCATOR_64_SIGNATURE = 0x07064b50;
    public static final int END_OF_CENTRAL_DIRECTORY_SIGNATURE = 0x06054b50;

    protected void Init()  throws IOException, InvalidZipFormatException {
        if (bb == null) {
            doInit();
        }
    }

    protected FileChannel.MapMode getFileChannelMode() {
        return FileChannel.MapMode.READ_ONLY;
    }

    protected String getOpenMode() {
        return "r";
    }

    protected void doInit() throws IOException, InvalidZipFormatException {
            raf = new RandomAccessFile(thisFile, getOpenMode());
            fc = raf.getChannel();
            bb = mapFile(thisFile.length());
    }

    protected MappedByteBuffer mapFile(long length) throws IOException {
        MappedByteBuffer buff;
        buff = fc.map(getFileChannelMode(), 0, length);
        buff.order(ByteOrder.LITTLE_ENDIAN);
        return buff;
    }

    public void close() throws IOException {
        bb = null;
        if (fc != null) {
            fc.close();
        }

        if (raf != null) {
            raf.close();
        }
    }

    public void visitFile(ZipVisitor visiter) throws IOException, InvalidZipFormatException {
        Init();

        Iterator    zipIter = iterator();

        while (zipIter.hasNext()) {
            ZipEntryImpl   next = (ZipEntryImpl) zipIter.next();
            visiter.startEntry(next);
            visiter.data(next.getBodyStream(bb));
            visiter.endEntry(next);
        }

    }

    Iterator    iterator() throws InvalidZipFormatException, IOException {
        Init();

        Iterator    retVal = null;

        try {
            retVal = new EntryIterator(bb);
        } catch (IllegalArgumentException iae) {
            retVal = new IndexIterator(bb, index);
        }

        return retVal;
    }


    public ZipEntry getEntry(String name) {
        try {
            Init();
            if (!index.isLoaded()) {
                index.load(bb);
            }
            return index.getEntry(name, bb);

        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    public InputStream getInputStream(ZipEntry zipEntry) throws IOException {
        Init();
        return zipEntry.getBodyStream(bb);
    }

    public String getName() {
        return thisFile.getName();
    }

    public class MyEnum implements Enumeration {
        public boolean hasMoreElements() {
            return false;
        }

        public Object nextElement() {
            return null;
        }
    }

    public Enumeration entries() {
        try {
            return (Enumeration) iterator();
        } catch (InvalidZipFormatException izf) {
            throw new RuntimeException(izf);
        } catch (IOException ioe) {
            throw new RuntimeException(ioe);
        }
    }

    public int size() {
        try {
            Init();
            if (!index.isLoaded())
                index.load(bb);
            return index.getCount();
        } catch (InvalidZipFormatException e) {
            return 0;
        } catch (IOException e) {
            return 0;
        }
    }

    protected void finalize() throws IOException {
        close();
    }


}
