package com.sourcelabs.zipper.zip;

import com.sourcelabs.zipper.EntryAlreadyExistsException;

import java.util.HashMap;
import java.nio.channels.FileChannel;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.io.IOException;

/**
 * Contains all the offset, size, etc. information that we need to reading/writing zip files.  This
 * maps very closely to all the information that is in the Central Directory of the zip.
 *
 * This class is also capable of writing it's entries out in the Central Directory format.
 *
 * <b>UNDONES</b>
 *   1)  Merge with {@link ZipIndex}.  I don't see any reason for this subclassing.  I had an idea
 * that I later decided was wrong.
 * @author willpugh  Apr 14, 2006 10:31:09 PM
 */
public class ZipIndexBase {

    public static final int HEADER_LEN = 46;

    protected int[]       offsets;
    protected int[]       compressedSizes;
    protected int[]       uncompressedSizes;
    protected String[]    names;
    protected HashMap     nameMap;
    protected HashMap     commentsMap;

    public void init(int count) {
        offsets = new int[count];
        compressedSizes = new int[count];
        uncompressedSizes = new int[count];
        names = new String[count];
        nameMap = new HashMap(count);
        commentsMap = null;
    }

    public int getCount() {
        return names.length;
    }

    public int getOffset(int i) {
        return offsets[i];
    }

    public void setOffset(int index, int offset) {
        offsets[index] = offset;
    }

    public int getCompressedSize(int index) {
        return compressedSizes[index];
    }

    public void setCompressedSize(int index, int size) {
        compressedSizes[index] = size;
    }

    public int getUncompressedSizes(int index) {
        return uncompressedSizes[index];
    }

    public void setUncompressedSizes(int index, int size) {
        uncompressedSizes[index] = size;
    }

    public String getName(int index) {
        return names[index];
    }

    public void setName(int index, String name) {
        names[index] = name;
        Integer i = (Integer) nameMap.put(name, new Integer(index));
        if (i != null) {
            nameMap.put(name, i);
            throw new EntryAlreadyExistsException(name + " already exists in index.");
        }
    }

    public boolean containsName(String name) {
        return nameMap.containsKey(name);
    }

    public void addComment(int index, String comment) {
        if (commentsMap == null)
            commentsMap = new HashMap();

        commentsMap.put(new Integer(index), comment);
    }

    public String getComment(int index) {
        if (commentsMap == null) {
            return null;
        }
        return (String) commentsMap.get(new Integer(index));
    }

    public String getComment(String name) {
        if (commentsMap == null) {
            return null;
        }

        Integer i = (Integer) nameMap.get(name);
        return (String) commentsMap.get(i);
    }

    public int writeToChannel(int index, FileChannel fc, ZipEntry entry) throws IOException {
        //UNDONE(willpugh) -- Deal with extra
        //UNDONE(willpugh) -- Deal with different encodings
        byte[]  rawName = names[index].getBytes();
        String  comment = getComment(index);
        byte[]  rawComment = null;
        int     commentLen = 0;
        if (comment != null) {
            rawComment = comment.getBytes();
            commentLen = rawComment.length;
        }
        byte[]  extra = entry.getExtra();
        int     extraLen = (extra != null ? extra.length : 0);

        int             headerLen = HEADER_LEN + rawName.length +
                                    commentLen + extraLen;
        ByteBuffer      bb = ByteBuffer.allocate(headerLen);
        bb.order(ByteOrder.LITTLE_ENDIAN);
        bb.putInt(ZipFile.CENTRAL_DIRECTORY_SIGNATURE);

        //UNDONE(willpugh) -- Fix this to report the right system
        bb.putShort((short)0);
        bb.putShort((short)20);
        bb.putShort((short)0);
        bb.putShort(entry.getCompression());
        bb.putInt(entry.getRawtime());
        bb.putInt(entry.getCrc_32());
        bb.putInt(entry.getCompressedSize());
        bb.putInt((int) entry.getUncompressedSize());
        bb.putShort((short) rawName.length);
        bb.putShort((short)extraLen);
        bb.putShort((short) commentLen);
        bb.putShort((short)0);
        bb.putShort((short)0);
        bb.putInt((short)0);
        bb.putInt(getOffset(index));
        bb.put(rawName);
        if (extra != null)
            bb.put(extra);
        if (comment != null)
            bb.put(rawComment);

        bb.position(0);
        fc.write(bb);
        return headerLen;
    }

}
